"""
Create a chip (subimage) of a SICD image.

For a basic help on the command-line, check

>>> python -m sarpy.utils.chip_sicd --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "John Gorman"

import os
import argparse
from typing import Tuple
import logging

import sarpy
from sarpy.io.complex.converter import conversion_utility
from sarpy.io.complex.sicd import SICDReader


def _verify_limits(limits):
    """
    Helper function to verify that the row/column limits are sensible.

    Parameters
    ----------
    limits : None|tuple|list

    Returns
    -------
    None|tuple
    """

    if limits is None:
        return limits

    temp_limits = [int(entry) for entry in limits]
    if len(temp_limits) != 2:
        raise ValueError('Got unexpected limits `{}`'.format(limits))
    if not (0 <= temp_limits[0] < temp_limits[1]):
        raise ValueError('Got unexpected limits `{}`'.format(limits))
    return temp_limits[0], temp_limits[1]


def create_chip(input_reader, out_directory, output_file=None, row_limits=None, col_limits=None,
                check_existence=True, check_older_version=False, preserve_nitf_information=False):
    """
    Create a chip of the given sicd file. At least one of `row_limits` and
    `col_limits` must be provided.

    Parameters
    ----------
    input_reader : str|SICDReader
    out_directory : str
        The directory of the output.
    output_file : None|str
        If `None`, then the name will mirror the original with row/col details appended.
    row_limits : None|Tuple[int, int]
        The limits for the rows, relative to this actual image, to be included.
    col_limits : None|Tuple[int, int]
        The limits for the columns, relative to this actual image, to be included.
    check_existence : bool
        Check for the existence of the file before overwriting?
    check_older_version : bool
        Try to use a less recent version of SICD (1.1), for possible application compliance issues?
    preserve_nitf_information : bool
        Try to preserve some of the original NITF information?
    """

    def get_suffix(limits, shift):
        if limits is None:
            return 'all'
        else:
            return '{0:d}-{1:d}'.format(shift+limits[0], shift+limits[1])

    if isinstance(input_reader, str):
        input_reader = SICDReader(input_reader)
    if not isinstance(input_reader, SICDReader):
        raise TypeError('We require that the input is a SICD reader or path to a sicd file.')

    row_limits = _verify_limits(row_limits)
    col_limits = _verify_limits(col_limits)

    if row_limits is None and col_limits is None:
        raise ValueError('At least one of row_limits and col_limits must be provided.')

    if output_file is None:
        fname = os.path.split(input_reader.file_name)[1]
        fstem, fext = os.path.splitext(fname)
        fstem += '_{}_{}'.format(
            get_suffix(row_limits, input_reader.sicd_meta.ImageData.FirstRow),
            get_suffix(col_limits, input_reader.sicd_meta.ImageData.FirstCol))
        fname = fstem + fext
    else:
        fname = os.path.split(output_file)[1]

    conversion_utility(
        input_reader, out_directory, output_files=fname,
        row_limits=row_limits, column_limits=col_limits,
        check_existence=check_existence,
        check_older_version=check_older_version,
        preserve_nitf_information=preserve_nitf_information)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Subset SICD file.",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input sicd data file.')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory. This directory MUST exist.')
    parser.add_argument(
        '-o', '--output_file', default=None, type=str)
    parser.add_argument(
        '-r', '--row_lims', default=None, nargs=2, type=int,
        help='Row limits for chip, integers: row_start row_stop')
    parser.add_argument(
        '-c', '--col_lims', default=None, nargs=2, type=int,
        help='Column limits for chip, integers: col_start col_stop')
    parser.add_argument(
        '-p', '--preserve', action='store_true',
        help='Try to preserve any NITF information?\n'
             'This only applies in the event that the file being read is a NITF')
    parser.add_argument(
        '-w', '--overwrite', action='store_true',
        help='Overwrite output file, if it already exists?')
    parser.add_argument(
        '--older', action='store_true',
        help='Try to use a less recent version of SICD (1.1),\n'
             'for possible application compliance issues?')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args()

    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    create_chip(
        args.input_file, args.output_directory, output_file=args.output_file,
        row_limits=args.row_lims, col_limits=args.col_lims,
        check_existence=not args.overwrite,
        check_older_version=args.older,
        preserve_nitf_information=args.preserve)



"""
The Compensated Phase History Data 0.3 definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union

from sarpy.io.phase_history.cphd1_elements.base import DEFAULT_STRICT

from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType
from sarpy.io.phase_history.cphd1_elements.CPHD import CPHDHeaderBase
from sarpy.io.phase_history.cphd0_3_elements.Data import DataType
from sarpy.io.phase_history.cphd0_3_elements.Global import GlobalType
from sarpy.io.phase_history.cphd0_3_elements.Channel import ChannelType
from sarpy.io.phase_history.cphd0_3_elements.SRP import SRPTyp
from sarpy.io.phase_history.cphd0_3_elements.Antenna import AntennaType
from sarpy.io.phase_history.cphd0_3_elements.VectorParameters import VectorParametersType

from sarpy.io.xml.base import Serializable, parse_xml_from_string, parse_xml_from_file
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, StringDescriptor

#########
# Module variables
_CPHD_SPECIFICATION_VERSION = '0.3'
_CPHD_SPECIFICATION_DATE = '2011-04-15T00:00:00Z'
_CPHD_SPECIFICATION_NAMESPACE = 'urn:CPHD:0.3'


#########
# CPHD header object

class CPHDHeader(CPHDHeaderBase):
    _fields = (
        'XML_DATA_SIZE', 'XML_BYTE_OFFSET', 'VB_DATA_SIZE', 'VB_BYTE_OFFSET',
        'CPHD_DATA_SIZE', 'CPHD_BYTE_OFFSET', 'CLASSIFICATION', 'RELEASE_INFO')
    _required = (
        'XML_DATA_SIZE', 'XML_BYTE_OFFSET', 'VB_DATA_SIZE', 'VB_BYTE_OFFSET',
        'CPHD_DATA_SIZE', 'CPHD_BYTE_OFFSET')
    # descriptor
    XML_DATA_SIZE = IntegerDescriptor(
        'XML_DATA_SIZE', _required, strict=True,
        docstring='Size of the XML Metadata in bytes. Does not include the 2 bytes '
                  'of the section terminator.')  # type: int
    XML_BYTE_OFFSET = IntegerDescriptor(
        'XML_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the XML Metadata in bytes.')  # type: int
    VB_DATA_SIZE = IntegerDescriptor(
        'VB_DATA_SIZE', _required, strict=True,
        docstring='Size of the Vector Based Metadata in bytes.')  # type: int
    VB_BYTE_OFFSET = IntegerDescriptor(
        'VB_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the Vector Based Metadata in bytes.')  # type: int
    CPHD_DATA_SIZE = IntegerDescriptor(
        'CPHD_DATA_SIZE', _required, strict=True,
        docstring='Size of the Compensated PHD arrays in bytes.')  # type: int
    CPHD_BYTE_OFFSET = IntegerDescriptor(
        'CPHD_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the CPHD data in bytes.')  # type: int
    CLASSIFICATION = StringDescriptor(
        'CLASSIFICATION', _required, strict=True, default_value='UNCLASSIFIED',
        docstring='Product classification information that is the human-readable banner.')  # type: str
    RELEASE_INFO = StringDescriptor(
        'RELEASE_INFO', _required, strict=True, default_value='UNRESTRICTED',
        docstring='Product release information.')  # type: str

    def __init__(self, XML_DATA_SIZE=None, XML_BYTE_OFFSET=None,
                 VB_DATA_SIZE=None, VB_BYTE_OFFSET=None,
                 CPHD_DATA_SIZE=None, CPHD_BYTE_OFFSET=None,
                 CLASSIFICATION='UNCLASSIFIED', RELEASE_INFO='UNRESTRICTED'):
        self.XML_DATA_SIZE = XML_DATA_SIZE
        self.XML_BYTE_OFFSET = XML_BYTE_OFFSET
        self.VB_DATA_SIZE = VB_DATA_SIZE
        self.VB_BYTE_OFFSET = VB_BYTE_OFFSET
        self.CPHD_DATA_SIZE = CPHD_DATA_SIZE
        self.CPHD_BYTE_OFFSET = CPHD_BYTE_OFFSET
        self.CLASSIFICATION = CLASSIFICATION
        self.RELEASE_INFO = RELEASE_INFO
        super(CPHDHeader, self).__init__()


class CPHDType(Serializable):
    """
    """

    _fields = (
        'CollectionInfo', 'Data', 'Global', 'Channel', 'SRP', 'RadarCollection', 'Antenna',
        'VectorParameters')
    _required = (
        'CollectionInfo', 'Data', 'Global', 'Channel', 'SRP', 'VectorParameters')
    # descriptors
    CollectionInfo = SerializableDescriptor(
        'CollectionInfo', CollectionInfoType, _required, strict=DEFAULT_STRICT,
        docstring='General information about the collection.')  # type: CollectionInfoType
    Data = SerializableDescriptor(
        'Data', DataType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe binary data components contained in the '
                  'product.')  # type: DataType
    Global = SerializableDescriptor(
        'Global', GlobalType, _required, strict=DEFAULT_STRICT,
        docstring='Global parameters that apply to metadata components and CPHD '
                  'signal arrays.')  # type: GlobalType
    Channel = SerializableDescriptor(
        'Channel', ChannelType, _required, strict=DEFAULT_STRICT,
        docstring='Channel specific parameters for CPHD channels.')  # type: ChannelType
    SRP = SerializableDescriptor(
        'SRP', SRPTyp, _required, strict=DEFAULT_STRICT,
        docstring='The Stabilization Reference Point (SRP) parameters.')  # type: SRPTyp
    RadarCollection = SerializableDescriptor(
        'RadarCollection', RadarCollectionType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, RadarCollectionType]
    Antenna = SerializableDescriptor(
        'Antenna', AntennaType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna parameters that describe antenna orientation, mainlobe '
                  'steering and gain patterns vs. '
                  'time.')  # type: Union[None, AntennaType]
    VectorParameters = SerializableDescriptor(
        'VectorParameters', VectorParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Structure specifying the Vector parameters provided for '
                  'each channel of a given product.')  # type: VectorParametersType

    def __init__(self, CollectionInfo=None, Data=None, Global=None, Channel=None,
                 SRP=None, RadarCollection=None, Antenna=None, VectorParameters=None, **kwargs):
        """

        Parameters
        ----------
        CollectionInfo : CollectionInfoType
        Data : DataType
        Global : GlobalType
        Channel : ChannelType
        SRP : SRPTyp
        RadarCollection : None|RadarCollectionType
        Antenna : None|AntennaType
        VectorParameters : VectorParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectionInfo = CollectionInfo
        self.Data = Data
        self.Global = Global
        self.Channel = Channel
        self.SRP = SRP
        self.RadarCollection = RadarCollection
        self.Antenna = Antenna
        self.VectorParameters = VectorParameters
        super(CPHDType, self).__init__(**kwargs)

    def to_xml_bytes(self, urn=None, tag='CPHD', check_validity=False, strict=DEFAULT_STRICT):
        if urn is None:
            urn = _CPHD_SPECIFICATION_NAMESPACE
        return super(CPHDType, self).to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict)

    def to_xml_string(self, urn=None, tag='CPHD', check_validity=False, strict=DEFAULT_STRICT):
        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')

    def get_pvp_dtype(self):
        """
        Gets the dtype for the corresponding PVP structured array. Note that they
        must all have homogeneous dtype.

        Returns
        -------
        numpy.dtype
            This will be a compound dtype for a structured array.
        """

        if self.VectorParameters is None:
            raise ValueError('No VectorParameters defined.')
        return self.VectorParameters.get_vector_dtype()

    @classmethod
    def from_xml_file(cls, file_path):
        """
        Construct the cphd object from a stand-alone xml file path.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        CPHDType
        """

        root_node, xml_ns = parse_xml_from_file(file_path)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Construct the cphd object from an xml string.

        Parameters
        ----------
        xml_string : str|bytes

        Returns
        -------
        CPHDType
        """

        root_node, xml_ns = parse_xml_from_string(xml_string)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)



"""
This module contains the base objects for use in the SICD elements, and the base serializable functionality.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Dict, Tuple, Optional

import numpy

from sarpy.io.xml.base import SerializableArray, create_new_node
from sarpy.io.xml.descriptors import BasicDescriptor


logger = logging.getLogger(__name__)
DEFAULT_STRICT = False
FLOAT_FORMAT = '0.17G'


class SerializableCPArrayDescriptor(BasicDescriptor):
    """A descriptor for properties of a list or array of specified extension of Serializable"""
    minimum_length = 4
    maximum_length = 4

    def __init__(
            self,
            name: str,
            child_type,
            tag_dict: Dict,
            required: Tuple[str, ...],
            strict: bool = DEFAULT_STRICT,
            docstring: Optional[str] = None):
        self.child_type = child_type
        tags = tag_dict[name]
        self.array = tags.get('array', False)
        if not self.array:
            raise ValueError(
                'Attribute {} is populated in the `_collection_tags` dictionary without `array`=True. '
                'This is inconsistent with using SerializableCPArrayDescriptor.'.format(name))

        self.child_tag = tags['child_tag']
        self._typ_string = 'numpy.ndarray[{}]:'.format(str(child_type).strip().split('.')[-1][:-2])

        super(SerializableCPArrayDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(SerializableCPArrayDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, SerializableCPArray):
            self.data[instance] = value
        else:
            xml_ns = getattr(instance, '_xml_ns', None)
            # noinspection PyProtectedMember, PyUnresolvedReferences
            if hasattr(instance, '_child_xml_ns_key') and self.name in instance._child_xml_ns_key:
                # noinspection PyProtectedMember
                xml_ns_key = instance._child_xml_ns_key[self.name]
            else:
                xml_ns_key = getattr(instance, '_xml_ns_key', None)

            the_inst = self.data.get(instance, None)
            if the_inst is None:
                self.data[instance] = SerializableCPArray(
                    coords=value, name=self.name, child_tag=self.child_tag,
                    child_type=self.child_type, _xml_ns=xml_ns, _xml_ns_key=xml_ns_key)
            else:
                the_inst.set_array(value)


class SerializableCPArray(SerializableArray):
    __slots__ = (
        '_child_tag', '_child_type', '_array', '_name', '_minimum_length',
        '_maximum_length', '_index_as_string', '_xml_ns', '_xml_ns_key')

    def __init__(
            self,
            coords=None,
            name: str = None,
            child_tag: str = None,
            child_type=None,
            _xml_ns: Optional[Dict[str, str]] = None,
            _xml_ns_key: Optional[str] = None):
        if hasattr(child_type, '_CORNER_VALUES'):
            self._index_as_string = True
        else:
            self._index_as_string = False
        super(SerializableCPArray, self).__init__(
            coords=coords, name=name, child_tag=child_tag, child_type=child_type,
            _xml_ns=_xml_ns, _xml_ns_key=_xml_ns_key)
        self._minimum_length = 4
        self._maximum_length = 4

    @property
    def FRFC(self) -> Optional[numpy.ndarray]:
        if self._array is None:
            return None
        return self._array[0].get_array()

    @property
    def FRLC(self) -> Optional[numpy.ndarray]:
        if self._array is None:
            return None
        return self._array[1].get_array()

    @property
    def LRLC(self) -> Optional[numpy.ndarray]:
        if self._array is None:
            return None
        return self._array[2].get_array()

    @property
    def LRFC(self) -> Optional[numpy.ndarray]:
        if self._array is None:
            return None
        return self._array[3].get_array()

    def _check_indices(self):
        if not self._index_as_string:
            self._array[0].index = 1
            self._array[1].index = 2
            self._array[2].index = 3
            self._array[3].index = 4
        else:
            self._array[0].index = '1:FRFC'
            self._array[1].index = '2:FRLC'
            self._array[2].index = '3:LRLC'
            self._array[3].index = '4:LRFC'

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT):
        if self.size == 0:
            return None  # nothing to be done
        if ns_key is None:
            anode = create_new_node(doc, tag, parent=parent)
        else:
            anode = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)
        for i, entry in enumerate(self._array):
            entry.to_node(doc, self._child_tag, ns_key=ns_key, parent=anode,
                          check_validity=check_validity, strict=strict)
        return anode



"""
The Global type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")

from typing import Union

import numpy

from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import FloatDescriptor, DateTimeDescriptor, \
    SerializableDescriptor
from sarpy.io.phase_history.cphd1_elements.Global import FxBandType, TropoParametersType, \
    IonoParametersType

from .base import DEFAULT_STRICT, FLOAT_FORMAT


class TimelineType(Serializable):
    """
    Parameters that describe the collection times for the data contained in the product.
    """

    _fields = ('CollectionRefTime', 'RcvTime1', 'RcvTime2')
    _required = ('CollectionRefTime', 'RcvTime1', 'RcvTime2')
    _numeric_format = {'RcvTime1': FLOAT_FORMAT, 'RcvTime2': FLOAT_FORMAT}
    # descriptors
    CollectionRefTime = DateTimeDescriptor(
        'CollectionRefTime', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Collection Reference Time (t_CRT). Time reference used for all receive times'
                  ' and all transmit times. All times are specified in seconds relative to t_CRT'
                  ' (i.e., t_CRT is slow time t = 0).')  # type: numpy.datetime64
    RcvTime1 = FloatDescriptor(
        'RcvTime1', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Earliest RcvTime value for any signal vector in the product.'
                  ' Time relative to Collection Reference Time.')  # type: float
    RcvTime2 = FloatDescriptor(
        'RcvTime2', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Latest RcvTime value for any signal vector in the product.'
                  ' Time relative to Collection Reference Time.')  # type: float

    def __init__(self, CollectionRefTime=None, RcvTime1=None, RcvTime2=None, **kwargs):
        """

        Parameters
        ----------
        CollectionRefTime : numpy.datetime64|datetime|date|str
        RcvTime1 : float
        RcvTime2 : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectionRefTime = CollectionRefTime
        self.RcvTime1 = RcvTime1
        self.RcvTime2 = RcvTime2
        super(TimelineType, self).__init__(**kwargs)


class FrcvBandType(Serializable, Arrayable):
    """
    Parameters that describe the received frequency limits for the signal array(s) contained in the product.
    """

    _fields = ('FrcvMin', 'FrcvMax')
    _required = _fields
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptors
    FrcvMin = FloatDescriptor(
        'FrcvMin', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Minimum frcv_1 PVP value for any signal vector in the product.')  # type: float
    FrcvMax = FloatDescriptor(
        'FrcvMax', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Maximum frcv_2 PVP value for any signal vector in the product.')  # type: float

    def __init__(self, FrcvMin=None, FrcvMax=None, **kwargs):
        """

        Parameters
        ----------
        FrcvMin : float
        FrcvMax : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.FrcvMin = FrcvMin
        self.FrcvMax = FrcvMax
        super(FrcvBandType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        return numpy.array([self.FrcvMin, self.FrcvMax], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        # type: (Union[numpy.ndarray, list, tuple]) -> FrcvBandType
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(FrcvMin=array[0], FrcvMax=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class GlobalType(Serializable):
    """
    The Global type definition.
    """

    _fields = ('Timeline', 'FrcvBand', 'FxBand', 'TropoParameters', 'IonoParameters')
    _required = ('Timeline', 'FrcvBand')
    # descriptors
    Timeline = SerializableDescriptor(
        'Timeline', TimelineType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the collection times for the data contained '
                  'in the product')  # type: TimelineType
    FrcvBand = SerializableDescriptor(
        'FrcvBand', FrcvBandType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the Frcv frequency limits for the signal array(s) '
                  'contained in the product.')  # type: FrcvBandType
    FxBand = SerializableDescriptor(
        'FxBand', FxBandType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the FX frequency limits for the signal array(s) '
                  'contained in the product.')  # type: FxBandType
    TropoParameters = SerializableDescriptor(
        'TropoParameters', TropoParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters used to compute the propagation delay due to the '
                  'troposphere.')  # type: Union[None, TropoParametersType]
    IonoParameters = SerializableDescriptor(
        'IonoParameters', IonoParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters used to compute propagation effects due to the '
                  'ionosphere.')  # type: Union[None, IonoParametersType]

    def __init__(self, Timeline=None, FrcvBand=None, FxBand=None,
                 TropoParameters=None, IonoParameters=None, **kwargs):
        """

        Parameters
        ----------
        Timeline : TimelineType
        FrcvBand : FrcvBandType|numpy.ndarray|list|tuple
        FxBand : None|FxBandType|numpy.ndarray|list|tuple
        TropoParameters : None|TropoParametersType
        IonoParameters : None|IonoParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Timeline = Timeline
        self.FrcvBand = FrcvBand
        self.FxBand = FxBand
        self.TropoParameters = TropoParameters
        self.IonoParameters = IonoParameters
        super(GlobalType, self).__init__(**kwargs)



"""
The ImageCreation elements.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Optional, Union

import numpy

from datetime import datetime, date

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringDescriptor, DateTimeDescriptor

from .base import DEFAULT_STRICT


class ImageCreationType(Serializable):
    """
    General information about the image creation.
    """

    _fields = ('Application', 'DateTime', 'Site', 'Profile')
    _required = ()
    # descriptors
    Application = StringDescriptor(
        'Application', _required, strict=DEFAULT_STRICT,
        docstring='Name and version of the application used to create the image.')  # type: str
    DateTime = DateTimeDescriptor(
        'DateTime', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Date and time the image creation application processed the image (UTC).')  # type: numpy.datetime64
    Site = StringDescriptor(
        'Site', _required, strict=DEFAULT_STRICT,
        docstring='The creation site of this SICD product.')  # type: str
    Profile = StringDescriptor(
        'Profile', _required, strict=DEFAULT_STRICT,
        docstring='Identifies what profile was used to create this SICD product.')  # type: str

    def __init__(
            self,
            Application: Optional[str] = None,
            DateTime: Union[None, numpy.datetime64, datetime, date, str] = None,
            Site: Optional[str] = None,
            Profile: Optional[str] = None,
            **kwargs):
        """

        Parameters
        ----------
        Application : str
        DateTime : numpy.datetime64|datetime|date|str
        Site : str
        Profile : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Application = Application
        self.DateTime = DateTime
        self.Site = Site
        self.Profile = Profile
        super(ImageCreationType, self).__init__(**kwargs)



"""
The AntennaType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, Optional

import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import BooleanDescriptor, FloatDescriptor, \
    SerializableDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import Poly1DType, XYZPolyType, GainPhasePolyType


class EBType(Serializable):
    """
    Electrical boresight (EB) steering directions for an electronically steered array.
    """

    _fields = ('DCXPoly', 'DCYPoly')
    _required = _fields
    # descriptors
    DCXPoly = SerializableDescriptor(
        'DCXPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Electrical boresight steering *X-axis direction cosine (DCX)* as a function of '
                  'slow time ``(variable 1)``.')  # type: Poly1DType
    DCYPoly = SerializableDescriptor(
        'DCYPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Electrical boresight steering *Y-axis direction cosine (DCY)* as a function of '
                  'slow time ``(variable 1)``.')  # type: Poly1DType

    def __init__(
            self,
            DCXPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            DCYPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            **kwargs):
        """
        Parameters
        ----------
        DCXPoly : Poly1DType|numpy.ndarray|list|tuple
        DCYPoly : Poly1DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DCXPoly = DCXPoly
        self.DCYPoly = DCYPoly
        super(EBType, self).__init__(**kwargs)

    def __call__(
            self,
            t: Union[float, int, numpy.ndarray]):
        """
        Evaluate the polynomial at points `t`. This passes `t` straight through
        to :func:`polyval` of `numpy.polynomial.polynomial` for each of
        `DCXPoly,DCYPoly` components. If any of `DCXPoly,DCYPoly` is not populated,
        then `None` is returned.

        Parameters
        ----------
        t : float|int|numpy.ndarray
            The point(s) at which to evaluate.

        Returns
        -------
        None|numpy.ndarray
        """

        if self.DCXPoly is None or self.DCYPoly is None:
            return None
        return numpy.array([self.DCXPoly(t), self.DCYPoly(t)])


class AntParamType(Serializable):
    """
    The antenna parameters container.
    """

    _fields = (
        'XAxisPoly', 'YAxisPoly', 'FreqZero', 'EB', 'Array', 'Elem', 'GainBSPoly', 'EBFreqShift', 'MLFreqDilation')
    _required = ('XAxisPoly', 'YAxisPoly', 'FreqZero', 'Array')
    _numeric_format = {'FreqZero': FLOAT_FORMAT}
    # descriptors
    XAxisPoly = SerializableDescriptor(
        'XAxisPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna X-Axis unit vector in ECF coordinates as a function of time ``(variable 1)``.'
    )  # type: XYZPolyType
    YAxisPoly = SerializableDescriptor(
        'YAxisPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna Y-Axis unit vector in ECF coordinates as a function of time ``(variable 1)``.'
    )  # type: XYZPolyType
    FreqZero = FloatDescriptor(
        'FreqZero', _required, strict=DEFAULT_STRICT,
        docstring='RF frequency *(f0)* used to specify the array pattern and electrical boresight *(EB)* '
                  'steering direction cosines.')  # type: float
    EB = SerializableDescriptor(
        'EB', EBType, _required, strict=DEFAULT_STRICT,
        docstring='Electrical boresight *(EB)* steering directions for an electronically '
                  'steered array.')  # type: EBType
    Array = SerializableDescriptor(
        'Array', GainPhasePolyType, _required, strict=DEFAULT_STRICT,
        docstring='Array pattern polynomials that define the shape of the main-lobe.')  # type: GainPhasePolyType
    Elem = SerializableDescriptor(
        'Elem', GainPhasePolyType, _required, strict=DEFAULT_STRICT,
        docstring='Element array pattern polynomials for electronically steered arrays.')  # type: GainPhasePolyType
    GainBSPoly = SerializableDescriptor(
        'GainBSPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Gain polynomial *(in dB)* as a function of frequency for boresight *(BS)* at :math:`DCX=0, DCY=0`. '
                  'Frequency ratio :math:`(f-f0)/f0` is the input variable ``(variable 1)``, and the constant '
                  'coefficient is always `0.0`.')  # type: Poly1DType
    EBFreqShift = BooleanDescriptor(
        'EBFreqShift', _required, strict=DEFAULT_STRICT,
        docstring="""
        Parameter indicating whether the electronic boresite shifts with frequency for an electronically steered array.
        
        * `False` - No shift with frequency.
        
        * `True` - Shift with frequency per ideal array theory.
        
        """)  # type: bool
    MLFreqDilation = BooleanDescriptor(
        'MLFreqDilation', _required, strict=DEFAULT_STRICT,
        docstring="""
        Parameter indicating the mainlobe (ML) width changes with frequency.
        
        * `False` - No change with frequency.
        
        * `True` - Change with frequency per ideal array theory.
        
        """)  # type: bool

    def __init__(
            self,
            XAxisPoly: XYZPolyType = None,
            YAxisPoly: XYZPolyType = None,
            FreqZero: float = None,
            EB: Optional[EBType] = None,
            Array: GainPhasePolyType = None,
            Elem: Optional[GainPhasePolyType] = None,
            GainBSPoly: Union[None, Poly1DType, numpy.ndarray, list, tuple] = None,
            EBFreqShift: Optional[bool] = None,
            MLFreqDilation: Optional[bool] = None,
            **kwargs):
        """
        Parameters
        ----------
        XAxisPoly : XYZPolyType
        YAxisPoly : XYZPolyType
        FreqZero : float
        EB : EBType
        Array : GainPhasePolyType
        Elem : GainPhasePolyType
        GainBSPoly : Poly1DType|numpy.ndarray|list|tuple
        EBFreqShift : bool
        MLFreqDilation : bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.XAxisPoly, self.YAxisPoly = XAxisPoly, YAxisPoly
        self.FreqZero = FreqZero
        self.EB = EB
        self.Array, self.Elem = Array, Elem
        self.GainBSPoly = GainBSPoly
        self.EBFreqShift, self.MLFreqDilation = EBFreqShift, MLFreqDilation
        super(AntParamType, self).__init__(**kwargs)

    def _apply_reference_frequency(self, reference_frequency):
        if self.FreqZero is not None:
            self.FreqZero += reference_frequency


class AntennaType(Serializable):
    """Parameters that describe the antenna illumination patterns during the collection."""
    _fields = ('Tx', 'Rcv', 'TwoWay')
    _required = ()
    # descriptors
    Tx = SerializableDescriptor(
        'Tx', AntParamType, _required, strict=DEFAULT_STRICT,
        docstring='The transmit antenna parameters.')  # type: AntParamType
    Rcv = SerializableDescriptor(
        'Rcv', AntParamType, _required, strict=DEFAULT_STRICT,
        docstring='The receive antenna parameters.')  # type: AntParamType
    TwoWay = SerializableDescriptor(
        'TwoWay', AntParamType, _required, strict=DEFAULT_STRICT,
        docstring='The bidirectional transmit/receive antenna parameters.')  # type: AntParamType

    def __init__(
            self,
            Tx: Optional[AntParamType] = None,
            Rcv: Optional[AntParamType] = None,
            TwoWay: Optional[AntParamType] = None,
            **kwargs):
        """

        Parameters
        ----------
        Tx : AntParamType
        Rcv : AntParamType
        TwoWay : AntParamType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Tx, self.Rcv, self.TwoWay = Tx, Rcv, TwoWay
        super(AntennaType, self).__init__(**kwargs)

    def _apply_reference_frequency(self, reference_frequency):
        """
        If the reference frequency is used, adjust the necessary fields accordingly.
        Expected to be called by SICD parent.

        Parameters
        ----------
        reference_frequency : float
            The reference frequency.

        Returns
        -------
        None
        """

        if self.Tx is not None:
            # noinspection PyProtectedMember
            self.Tx._apply_reference_frequency(reference_frequency)
        if self.Rcv is not None:
            # noinspection PyProtectedMember
            self.Rcv._apply_reference_frequency(reference_frequency)
        if self.TwoWay is not None:
            # noinspection PyProtectedMember
            self.TwoWay._apply_reference_frequency(reference_frequency)



"""
Apply or remove an impulse response side lobe control window function.

For a basic help on the command-line, check

>>> python -m sarpy.utils.sicd_sidelobe_control --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Valkyrie Systems Corporation"

import argparse
import logging

import numpy as np

from sarpy.io.complex.sicd import SICDReader, SICDWriter, validate_sicd_for_writing
from sarpy.processing.sicd.spectral_taper import Taper, apply_spectral_taper


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Remove and/or apply a sidelobe control spectral taper window to a SICD file.",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file', help='Path to the input data SICD file.')
    parser.add_argument(
        'output_file', metavar='output_file', help='Path to the output SICD file.\n')
    window_args_parser(parser)
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args(args)

    window = 'UNIFORM' if args.window is None else args.window.upper()
    default_pars = Taper().default_pars

    # Convert the list of par values into a dict of pars {name: value}.
    pars = {p: args.pars[n] for n, p in enumerate(default_pars.get(window, {}))}

    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    reader = SICDReader(args.input_file)

    taper = Taper(window, pars)

    cdata, mdata = apply_spectral_taper(reader, taper)

    # Write the results to a new SICD file.
    writer = SICDWriter(
        file_object=args.output_file,
        sicd_meta=validate_sicd_for_writing(mdata),
        check_existence=False
    )
    writer.write(cdata.astype(np.complex64))
    writer.close()


def window_args_parser(parser):
    def pars_text(pars):
        prefix = 'takes 1 parameter: ' if len(pars) == 1 else f'takes {len(pars)} parameters: '
        return prefix + ', '.join([f'"{p.lower()}"' for p in pars.keys()])

    window_name_choices = [w.lower() for w in Taper().default_pars.keys()]
    window_pars_choices = [f'  "{w.lower()}": {pars_text(p)}' for w, p in Taper().default_pars.items() if p]
    parser.add_argument(
        '-w', '--window', type=str.lower, choices=window_name_choices,
        help='The name of the window function.  Acceptable (case insensitive) names are:\n'
             '  "uniform": This is a flat spectral taper (i.e., no spectral taper).\n'
             '  "hamming": Hamming window taper (0.54 + 0.46 * cos(2*pi*n/M))\n'
             '  "hanning" | "hann":  Hann window taper (0.5 + 0.5 * cos(2*pi*n/M))\n'
             '  "general_hamming": Raised cosine window (alpha + (1 - alpha) * cos(2*pi*n/M)), default (alpha = 0.5)\n'
             '  "kaiser": Kaiser-Bessel window (I0[beta * sqrt(1 - (2*n/M - 1)**2)] / I0[beta]) default (beta = 14)\n'
             '  "taylor": Taylor window: Default (nbar = 4, sll = -30)\n'
    )
    parser.add_argument(
        '-p', '--pars', nargs='+',
        help=("One or more parameter values to modify the window characteristics.\n" +
              '\n'.join(window_pars_choices +
                        ["For example:",
                         "  --window general_hamming --pars 0.6",
                         "  --window taylor          --pars 5 -35.0",
                         "  --window kaiser          --pars 15"]
                        )
              )
    )


if __name__ == '__main__':
    main()    # pragma: no cover



"""
Common utils for CPHD 1.0 functionality.
"""

import numpy

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Daniel Pressler, Valkyrie")


#########
# Module variables
_DTYPE_LOOKUP = {
    "U1": numpy.dtype('>u1'),
    "U2": numpy.dtype('>u2'),
    "U4": numpy.dtype('>u4'),
    "U8": numpy.dtype('>u8'),
    "I1": numpy.dtype('>i1'),
    "I2": numpy.dtype('>i2'),
    "I4": numpy.dtype('>i4'),
    "I8": numpy.dtype('>i8'),
    "F4": numpy.dtype('>f4'),
    "F8": numpy.dtype('>f8'),
    "CI2": numpy.dtype([('real', '>i1'), ('imag', '>i1')]),
    "CI4": numpy.dtype([('real', '>i2'), ('imag', '>i2')]),
    "CI8": numpy.dtype([('real', '>i4'), ('imag', '>i4')]),
    "CI16": numpy.dtype([('real', '>i8'), ('imag', '>i8')]),
    "CF8": numpy.dtype('>c8'),
    "CF16": numpy.dtype('>c16')}


def _single_binary_format_string_to_dtype(form):
    """
    Convert a CPHD datatype into a dtype.

    Parameters
    ----------
    form

    Returns
    -------
    numpy.dtype
    """

    if form.startswith('S'):
        return numpy.dtype(form)
    else:
        return _DTYPE_LOOKUP[form]


def binary_format_string_to_dtype(format_string):
    """
    Return the numpy.dtype for CPHD Binary Format string (table 10-2).

    Parameters
    ----
    format_string: str
        PVP type designator (e.g., :code:`'I1', 'I4', 'CF8'`, etc.).

    Returns
    -------
    numpy.dtype
        The equivalent `numpy.dtype` of the PVP format string
        (e.g., :code:`numpy.int8, numpy.int32, numpy.complex64`, etc.).
    """

    components = format_string.split(';')
    if '=' in components[0]:
        assert format_string.endswith(';'), 'Format strings describing multiple parameters must end with a semi-colon'
        comptypes = []
        for comp in components[:-1]:
            kvp = comp.split('=')
            comptypes.append((kvp[0], _single_binary_format_string_to_dtype(kvp[1])))

        keys, types = list(zip(*comptypes))
        if keys == ('X', 'Y', 'Z') and len(set(types)) == 1:
            # special handling of XYZ types
            dtype = numpy.dtype((comptypes[0][1], 3))
        elif keys == ('DCX', 'DCY') and len(set(types)) == 1:
            # special handling of DCX/DCY types
            dtype = numpy.dtype((comptypes[0][1], 2))
        else:
            dtype = numpy.dtype(comptypes)
    else:
        dtype = _single_binary_format_string_to_dtype(components[0])
    return dtype


def homogeneous_dtype(format_string, return_length=False):
    """
    Determine a numpy.dtype (including endianness) from a CPHD format string, requiring
    that any multiple parts are all identical.

    Parameters
    ----------
    format_string : str
    return_length : bool
        Return the number of elements?

    Returns
    -------
    numpy.dtype|(numpy.dtype, int)
        Tuple of (`numpy.dtype`, # of elements) if `return_length`, `numpy.dtype` otherwise

    """

    raw_dtype = binary_format_string_to_dtype(format_string)
    if raw_dtype.names is None:
        # Simple or subarray
        dtype = raw_dtype.base
        num_elements = max(1, sum(raw_dtype.shape))
    else:
        # Structured
        dtype_set = {v[0] for v in raw_dtype.fields.values()}
        if len(dtype_set) == 1:
            dtype = dtype_set.pop()
            num_elements = len(raw_dtype.names)
        else:
            raise ValueError("Format string {} was heterogeneous (dtype_set={})".format(format_string, dtype_set))

    if return_length:
        return dtype, num_elements
    else:
        return dtype



"""
Module for converting between various latitude/longitude representations.
"""

import re

import numpy


__classification__ = "UNCLASSIFIED"
__author__ = "Wade Schwartzkopf"


def string(value, latlon, num_units=3, precision=None, delimiter='',
           include_symbols=True, signed=False, padded=True):
    """
    Convert latitude/longitude numeric values to customizable string format.

    Supports ISO 6709:2008 formatted geographic coordinates:

    * Annex D (human interface)
        delimiter = ''; include_symbols = true; padded = true; signed = false
    * Annex H (string representation)
        delimiter = ''; include_symbols = false; padded = true; signed = true

    Parameters
    ----------
    value : float|numpy.ndarray|list|tuple
        Value of latitude or longitude in decimal degrees or dms vector.
    latlon : str
        One of {'lat', 'lon'}, required for formatting string.
    num_units : int
        1 - decimal degrees; 2 - degrees/minutes; 3 - degrees/minutes/seconds.
        Default is 3.
    delimiter : str|list|tuple
        Separators between degrees/minutes/seconds/hemisphere.  Default is '' (empty).
    include_symbols : bool
        Whether to include degree, minute, second symbols.  Default is true.
    signed : bool
        Whether to use +/- or N/S/E/W to represent hemisphere.
        Default is false (N/S/E/W).
    precision : int
        Number of decimal points shown in finest unit.  Default is 5 if
        num_units==1, otherwise 0.
    padded : bool
        Whether to use zeros to pad out to consistent string length (3 digits for
        longitude degrees, 2 digits for all other elements).  Default is true.
    """

    if isinstance(value, (numpy.ndarray, list, tuple)):
        value = num(value)
    elif not isinstance(value, float):
        value = float(value)
    # value should now be in decimal degrees

    # Precision.  Default is dependent on other input arguments.
    if precision is None:
        if num_units == 1:
            precision = 5
        else:
            precision = 0
    # Symbols
    if include_symbols:
        latlon_symbols = ('\xB0', "'", '"')
    else:
        latlon_symbols = ('', '', '')
    # Delimiters
    try:
        if len(delimiter) != 3:
            delimiter = [delimiter]*num_units
    except:  # Must be a scalar if len() didn't work
        delimiter = [delimiter]*num_units
    if signed:
        delimiter[num_units-1] = ''  # No separator needed for hemisphere
    # Differences between latitude and longitude
    if latlon == 'lat':
        if value > 0:
            hemisphere = 'N'
            latlon_sign = '+'
        else:
            hemisphere = 'S'
            latlon_sign = '-'
        degrees_digits = 2
    elif latlon == 'lon':
        if value > 180:
            value = value - 360
        if value > 0:
            hemisphere = 'E'
            latlon_sign = '+'
        else:
            hemisphere = 'W'
            latlon_sign = '-'
        degrees_digits = 3
    # Compute degrees/minutes/seconds
    new_value = abs(value)
    value = [None]*num_units
    for i in range(num_units):
        fraction = new_value % 1.
        value[i] = int(new_value)
        new_value = fraction*60
    value[-1] = value[-1] + fraction
    if num_units > 1 and round(value[-1], precision) == 60:  # Seconds of 60 is invalid
        value[-1] = 0 
        value[-2] = value[-2] + 1
        if num_units == 3 and value[-2] == 60:  # If adding 1 to minutes makes minutes 60 which is also invalid
            value[-2] = 0
            value[-3] = value[-3] + 1
    # Build string
    latlon_string = ''
    for i in range(num_units):
        if padded:
            if i == 0:
                int_digits = degrees_digits
            else:
                int_digits = 2  # True for all but longitude degrees
        else:
            int_digits = 1
        if (i + 1) == num_units:
            precision_digits = precision
        else:
            precision_digits = 0
        if precision_digits > 0:
            int_digits = int_digits + 1  # Account for the decimal point
        latlon_string = '%s%0*.*f%s%s' % \
                        (latlon_string, int_digits + precision_digits, precision_digits,
                         abs(value[i]), latlon_symbols[i], delimiter[i])
    if signed:
        latlon_string = latlon_sign + latlon_string
    else:
        latlon_string = latlon_string + hemisphere
    return latlon_string


def dms(degrees):
    """
    Calculate degrees, minutes, seconds representation from decimal degrees.

    Parameters
    ----------
    degrees : float

    Returns
    -------
    (int, int, float)
    """

    degrees_int = int(abs(degrees))	 # integer degrees
    degrees_frac = abs(degrees) - degrees_int  # fractional degrees, used to compute minutes
    minutes_int = float(int(degrees_frac * 60))  # integer minutes
    minutes_frac = degrees_frac * 60 - minutes_int  # fractional minutes, used to compute seconds
    seconds = minutes_frac * 60  # decimal seconds

    # Handle sign.  Degrees portion will contain the sign of the coordinate.
    # Minutes and seconds will always be positive.
    # sign function returns -1, 0, +1 for x < 0, x == 0, x > 0, respectively
    if degrees < 0:
        degrees_int *= -1

    return degrees_int, minutes_int, seconds


def num(latlon_input):
    """
    Convert a variety of lat/long formats into decimal degree value.

    This should handle any string compliant with the ISO 6709:2008 standard
    or any of a number of variants for describing lat/long coordinates.
    Also handles degree/minutes/seconds passed in as a tuple/list/array.

    Parameters
    ----------
    latlon_input : numpy.ndarray|list|tuple|str

    Returns
    -------
    float
    """

    # Vector format degrees/minutes/seconds
    if isinstance(latlon_input, (numpy.ndarray, list, tuple)) and len(latlon_input) == 3:
        return float(latlon_input[0]) + float(latlon_input[1])/60. + float(latlon_input[2])/3600.

    if not isinstance(latlon_input, str):
        raise ValueError('Expected a (degree, minutes, seconds) tuple of string. '
                         'Got type {}'.format(type(latlon_input)))
    # String input
    # Handles decimal degrees and degree/minutes/second with delimiters
    # Any non-numeric characters in string are considered delimiters
    tokens_str = list(filter(lambda x: len(x.strip()) > 0, re.split(r'[^.\d]', latlon_input)))
    tokens = [float(x) for x in tokens_str]
    decimal_degrees = numpy.polynomial.polynomial.polyval(1/60., numpy.abs(tokens))
    if ('W' in latlon_input or 'S' in latlon_input) != ('-' in latlon_input):
        decimal_degrees = -decimal_degrees
    # Handles degree/minutes/second with no delimiters DDD,DDDMM,DDDMMSS
    if len(tokens) == 1:
        for i in range(min(3, int(len(tokens_str[0].split('.')[0])/2)-1)):
            decimal_degrees = (numpy.fix(decimal_degrees/100) +
                               numpy.fmod(decimal_degrees, 100)/60.)
    # Error checking should occur here
    if (len(tokens) < 1 or len(tokens) > 3 or
       decimal_degrees < -180 or decimal_degrees > 360 or
       sum(c.isalpha() for c in latlon_input) > 1):
        decimal_degrees = float('nan')  # Unparseable inputs are returned as NaN
    return float(decimal_degrees)



"""
Provides coordinate transforms for WGS-84 and ECF coordinate systems
"""

import numpy

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Wade Schwartzkopf")

#####
# WGS-84 parameters and related derived parameters
_A = 6378137.0            # Semi-major radius (m)
_F = 1/298.257223563      # Flattening
_GM = 3986004.418E8       # Earth's gravitational constant (including atmosphere)
_W = 7292115.1467E-11     # Angular velocity (radians/second), not including precession
_B = _A - _F*_A           # 6356752.3142, Semi-minor radius (m)
_A2 = _A*_A
_B2 = _B*_B
_E2 = (_A2-_B2)/_A2  # 6.69437999014E-3, First eccentricity squared
_E4 = _E2*_E2
_OME2 = 1.0 - _E2
_EB2 = (_A2 - _B2)/_B2


def _validate(arr):
    if not isinstance(arr, numpy.ndarray):
        arr = numpy.array(arr, dtype='float64')

    if arr.shape[-1] != 3:
        raise ValueError(
            'The input argument should represent geographical coordinates, so the '
            'final dimension should have size 3. Got shape {}.'.format(arr.shape))
    orig_shape = arr.shape
    arr = numpy.reshape(arr, (-1, 3))  # this is just a view
    return arr, orig_shape


def ecf_to_geodetic(ecf, ordering='latlong'):
    """
    Converts ECF (Earth Centered Fixed) coordinates to WGS-84 coordinates.

    Parameters
    ----------
    ecf : numpy.ndarray|list|tuple
    ordering : str
        If 'longlat', then the return will be `[longitude, latitude, hae]`.
        Otherwise, the return will be `[latitude, longitude, hae]`.

    Returns
    -------
    numpy.ndarray
        The WGS-84 coordinates, of the same shape as `ecf`.
    """

    ecf, orig_shape = _validate(ecf)

    x = ecf[:, 0]
    y = ecf[:, 1]
    z = ecf[:, 2]

    llh = numpy.full(ecf.shape, numpy.nan, dtype=numpy.float64)

    r = numpy.sqrt((x * x) + (y * y))

    # Check for invalid solution
    valid = ((_A*r)*(_A*r) + (_B*z)*(_B*z) > (_A2 - _B2)*(_A2 - _B2))

    # calculate intermediates
    F = 54.0*_B2*z*z  # not the WGS 84 flattening parameter
    G = r*r + _OME2*z*z - _E2*(_A2 - _B2)
    C = _E4*F*r*r/(G*G*G)
    S = (1.0 + C + numpy.sqrt(C*C + 2*C))**(1./3)
    P = F/(3.0*(G*(S + 1.0/S + 1.0))**2)
    Q = numpy.sqrt(1.0 + 2.0*_E4*P)
    R0 = -P*_E2*r/(1.0 + Q) + numpy.sqrt(numpy.abs(0.5*_A2*(1.0 + 1/Q) - P*_OME2*z*z/(Q*(1.0 + Q)) - 0.5*P*r*r))
    T = r - _E2*R0
    U = numpy.sqrt(T*T + z*z)
    V = numpy.sqrt(T*T + _OME2*z*z)
    z0 = _B2*z/(_A*V)

    # account for ordering
    if ordering.lower() == 'longlat':
        inds = [0, 1, 2]
    else:
        inds = [1, 0, 2]
    # calculate longitude
    llh[valid, inds[0]] = numpy.rad2deg(numpy.arctan2(y[valid], x[valid]))
    # calculate latitude
    llh[valid, inds[1]] = numpy.rad2deg(numpy.arctan2(z[valid] + _EB2*z0[valid], r[valid]))
    # calculate altitude
    llh[valid, inds[2]] = U[valid]*(1.0 - _B2/(_A*V[valid]))
    return numpy.reshape(llh, orig_shape)


def geodetic_to_ecf(llh, ordering='latlong'):
    """
    Converts WGS-84 coordinates to ECF (Earth Centered Fixed).

    Parameters
    ----------
    llh : numpy.ndarray|list|tuple
    ordering : str
        If 'longlat', then the input is `[longitude, latitude, hae]`.
        Otherwise, the input is `[latitude, longitude, hae]`.

    Returns
    -------
    numpy.ndarray
        The ECF coordinates, of the same shape as `llh`.
    """

    llh, orig_shape = _validate(llh)

    # account for ordering
    if ordering.lower() == 'longlat':
        inds = [0, 1, 2]
    else:
        inds = [1, 0, 2]
    lon = llh[:, inds[0]]
    lat = llh[:, inds[1]]
    alt = llh[:, inds[2]]

    out = numpy.full(llh.shape, numpy.nan, dtype=numpy.float64)
    # calculate distance to surface of ellipsoid
    r = _A / numpy.sqrt(1.0 - _E2*numpy.sin(numpy.deg2rad(lat)) * numpy.sin(numpy.deg2rad(lat)))

    # calculate coordinates
    out[:, 0] = (r + alt)*numpy.cos(numpy.deg2rad(lat))*numpy.cos(numpy.deg2rad(lon))
    out[:, 1] = (r + alt)*numpy.cos(numpy.deg2rad(lat))*numpy.sin(numpy.deg2rad(lon))
    out[:, 2] = (r + alt - _E2*r)*numpy.sin(numpy.deg2rad(lat))
    return numpy.reshape(out, orig_shape)


def wgs_84_norm(ecf):
    """
    Calculates the normal vector to the WGS_84 ellipsoid at the given ECF coordinates.

    Parameters
    ----------
    ecf : numpy.ndarray|list|tuple

    Returns
    -------
    numpy.ndarray
        The normal vector, of the same shape as `ecf`.
    """

    ecf, orig_shape = _validate(ecf)
    out = numpy.copy(ecf)/numpy.array([_A2, _A2, _B2], dtype=numpy.float64)
    out = out/(numpy.linalg.norm(out, axis=1)[:, numpy.newaxis])
    return numpy.reshape(out, orig_shape)


def _ecf_to_ned_matrix(orp_coord):
    """
    Get the rotation matrix for converting ECF to NED coordinate system
    conversion.

    Note: The array orientation convention indicates array multiplication on the
    RIGHT, so this is the transpose of the transform matrix for left multiplication.

    Parameters
    ----------
    orp_coord : numpy.ndarray
        The origin reference point. This is assumed given in ECF coordinates.

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(orp_coord, numpy.ndarray) or orp_coord.ndim != 1 or orp_coord.size != 3:
        raise ValueError('orp_coord must be a one-dimensional array of length 3.')
    llh = ecf_to_geodetic(orp_coord)

    angle2 = numpy.deg2rad(-90 - llh[0])
    angle1 = numpy.deg2rad(llh[1])

    matrix1 = numpy.array([[numpy.cos(angle1), -numpy.sin(angle1), 0],
                           [numpy.sin(angle1), numpy.cos(angle1), 0],
                           [0, 0, 1]], dtype='float64')
    matrix2 = numpy.array([[numpy.cos(angle2), 0, numpy.sin(angle2)],
                           [0, 1, 0],
                           [-numpy.sin(angle2), 0, numpy.cos(angle2)]], dtype='float64')
    return matrix1.dot(matrix2)


def ecf_to_ned(ecf_coords, orp_coord, absolute_coords=True):
    """
    Convert from ECF to North-East-Down (NED) coordinates.

    Parameters
    ----------
    ecf_coords : numpy.ndarray
    orp_coord : numpy.ndarray
    absolute_coords : bool
        Are these absolute (i.e. position) coordinates? The alternative is relative
        coordinates like velocity, acceleration, or unit vector values.

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(orp_coord, numpy.ndarray):
        orp_coord = numpy.array(orp_coord, dtype='float64')
    transform = _ecf_to_ned_matrix(orp_coord)
    # NB: orp_coord is guaranteed to be shape (3, )
    ecf_coords, o_shape = _validate(ecf_coords)
    if absolute_coords:
        out = (ecf_coords - orp_coord).dot(transform)
    else:
        out = ecf_coords.dot(transform)
    return numpy.reshape(out, o_shape)


def ned_to_ecf(ned_coords, orp_coord, absolute_coords=True):
    """
    Convert from North-East-Down (NED) to ECF coordinates.

    Parameters
    ----------
    ned_coords : numpy.ndarray
        The NED coordinates.
    orp_coord : numpy.ndarray
        The Origin Reference Point in ECF coordinates.
    absolute_coords : bool
        Are these absolute (i.e. position) coordinates? The alternative is relative
        coordinates like velocity, acceleration, or unit vector values.

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(orp_coord, numpy.ndarray):
        orp_coord = numpy.array(orp_coord, dtype='float64')
    transform = _ecf_to_ned_matrix(orp_coord).transpose()  # transpose = inverse here
    # NB: orp_coord is guaranteed to be shape (3, )
    ned_coords, o_shape = _validate(ned_coords)

    out = ned_coords.dot(transform)
    if absolute_coords:
        out += orp_coord
    return numpy.reshape(out, o_shape)


def _ecf_to_enu_matrix(orp_coord):
    """
    Get the rotation matrix for converting from ECF to ENU.

    Note: The array orientation convention indicates array multiplication on the
    RIGHT, so this is the transpose of the transform matrix for left multiplication.

    Parameters
    ----------
    orp_coord : numpy.ndarray
        The origin reference point. This is assumed given in ECF coordinates.

    Returns
    -------
    numpy.ndarray
    """

    ned_matrix = _ecf_to_ned_matrix(orp_coord)
    ned_to_enu = numpy.array([[0, 1, 0], [1, 0, 0], [0, 0, -1]], dtype='float64')
    return ned_matrix.dot(ned_to_enu)


def ecf_to_enu(ecf_coords, orp_coord, absolute_coords=True):
    """
    Convert from ECF to East-North-Up (ENU) coordinates.

    Parameters
    ----------
    ecf_coords : numpy.ndarray
    orp_coord : numpy.ndarray
    absolute_coords : bool
        Are these absolute (i.e. position) coordinates? The alternative is relative
        coordinates like velocity, acceleration, or unit vector values.

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(orp_coord, numpy.ndarray):
        orp_coord = numpy.array(orp_coord, dtype='float64')
    transform = _ecf_to_enu_matrix(orp_coord)
    # NB: orp_coord is guaranteed to be shape (3, )
    ecf_coords, o_shape = _validate(ecf_coords)
    if absolute_coords:
        out = (ecf_coords - orp_coord).dot(transform)
    else:
        out = ecf_coords.dot(transform)
    return numpy.reshape(out, o_shape)


def enu_to_ecf(enu_coords, orp_coord, absolute_coords=True):
    """
    Convert from East-North-UP (ENU) to ECF coordinates.

    Parameters
    ----------
    enu_coords : numpy.ndarray
        The ENU coordinates.
    orp_coord : numpy.ndarray
        The Origin Reference Point in ECF coordinates.
    absolute_coords : bool
        Are these absolute (i.e. position) coordinates? The alternative is relative
        coordinates like velocity, acceleration, or unit vector values.

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(orp_coord, numpy.ndarray):
        orp_coord = numpy.array(orp_coord, dtype='float64')
    transform = _ecf_to_enu_matrix(orp_coord).transpose()  # transpose = inverse here
    # NB: orp_coord is guaranteed to be shape (3, )
    enu_coords, o_shape = _validate(enu_coords)

    out = enu_coords.dot(transform)
    if absolute_coords:
        out += orp_coord
    return numpy.reshape(out, o_shape)



"""
The MeasurementType definition for SIDD 1.0.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union

from sarpy.io.product.sidd2_elements.base import DEFAULT_STRICT
from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import SerializableDescriptor
from sarpy.io.product.sidd2_elements.blocks import RowColIntType, XYZPolyType
from sarpy.io.product.sidd2_elements.Measurement import PolynomialProjectionType, \
    GeographicProjectionType, PlaneProjectionType, CylindricalProjectionType


class MeasurementType(Serializable):
    """
    Geometric SAR information required for measurement/geolocation.
    """

    _fields = (
        'PolynomialProjection', 'GeographicProjection', 'PlaneProjection', 'CylindricalProjection',
        'PixelFootprint', 'ARPPoly')
    _required = ('PixelFootprint', 'ARPPoly')
    _choice = ({'required': False, 'collection': ('PolynomialProjection', 'GeographicProjection',
                                                  'PlaneProjection', 'CylindricalProjection')}, )
    # Descriptor
    PolynomialProjection = SerializableDescriptor(
        'PolynomialProjection', PolynomialProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial pixel to ground. Should only used for sensor systems where the radar '
                  'geometry parameters are not recorded.')  # type: Union[None, PolynomialProjectionType]
    GeographicProjection = SerializableDescriptor(
        'GeographicProjection', GeographicProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Geographic mapping of the pixel grid referred to as GGD in the '
                  'Design and Exploitation document.')  # type: Union[None, GeographicProjectionType]
    PlaneProjection = SerializableDescriptor(
        'PlaneProjection', PlaneProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Planar representation of the pixel grid referred to as PGD in the '
                  'Design and Exploitation document.')  # type: Union[None, PlaneProjectionType]
    CylindricalProjection = SerializableDescriptor(
        'CylindricalProjection', CylindricalProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Cylindrical mapping of the pixel grid referred to as CGD in the '
                  'Design and Exploitation document.')  # type: Union[None, CylindricalProjectionType]
    PixelFootprint = SerializableDescriptor(
        'PixelFootprint', RowColIntType, _required, strict=DEFAULT_STRICT,
        docstring='Size of the image in pixels.')  # type: RowColIntType
    ARPPoly = SerializableDescriptor(
        'ARPPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Center of aperture polynomial (units = m) based upon time into '
                  'the collect.')  # type: XYZPolyType

    def __init__(self, PolynomialProjection=None, GeographicProjection=None, PlaneProjection=None,
                 CylindricalProjection=None, ARPPoly=None, **kwargs):
        """

        Parameters
        ----------
        PolynomialProjection : PolynomialProjectionType
        GeographicProjection : GeographicProjectionType
        PlaneProjection : PlaneProjectionType
        CylindricalProjection : CylindricalProjectionType
        ARPPoly : XYZPolyType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PolynomialProjection = PolynomialProjection
        self.GeographicProjection = GeographicProjection
        self.PlaneProjection = PlaneProjection
        self.CylindricalProjection = CylindricalProjection
        self.ARPPoly = ARPPoly
        super(MeasurementType, self).__init__(**kwargs)

    @property
    def ProjectionType(self):
        """str: *READ ONLY* Identifies the specific image projection type supplied."""
        for attribute in self._choice[0]['collection']:
            if getattr(self, attribute) is not None:
                return attribute
        return None

    @property
    def ReferencePoint(self):
        """
        None|ReferencePointType: *READ ONLY* Gets the reference point.
        """
        
        if self.ProjectionType:
            return getattr(self, self.ProjectionType).ReferencePoint
        return None



"""
The ExploitationFeaturesType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import datetime
from typing import Union, List

import numpy

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, ParametersDescriptor, \
    FloatDescriptor, FloatModularDescriptor, StringDescriptor, StringRegexDescriptor, \
    DateTimeDescriptor, SerializableListDescriptor
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.geometry.geocoords import ecf_to_geodetic

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.product.sidd2_elements.ExploitationFeatures import ExploitationCalculator as BaseExploitationCalculator
from .blocks import AngleZeroToExclusive360MagnitudeType, RadarModeType, RangeAzimuthType, RowColDoubleType

from sarpy.io.product.sidd2_elements.ExploitationFeatures import (
    InputROIType,
    get_collection_resolution,
)

__REUSED__ = (
    InputROIType,
)


RCV_POLARIZATION_PATTERN = 'V|H|X|Y|S|E|RHC|LHC|UNKNOWN|OTHER[^:]*'
TX_POLARIZATION_PATTERN = RCV_POLARIZATION_PATTERN + '|SEQUENCE'
logger = logging.getLogger(__name__)

_sicd_type_text = 'Requires instance of SICDType,\n\tgot type `{}`'
_exp_calc_text = 'Requires input which is an instance of ExploitationCalculator,\n\tgot type `{}`'


class ExploitationCalculator(BaseExploitationCalculator):
    @property
    def Shadow(self):
        # type: () -> AngleZeroToExclusive360MagnitudeType
        """
        AngleMagnitudeType: The shadow angle and magnitude.
        """

        shadow = self.ETP - self.slant_x/(self.slant_x.dot(self.ETP))
        shadow_prime = shadow - (shadow.dot(self.normal_vector)/self.slant_z.dot(self.normal_vector))*self.slant_z
        theta_shadow = numpy.rad2deg(numpy.arctan2(self.row_vector.dot(shadow_prime), self.col_vector.dot(shadow_prime)))
        if theta_shadow < 0:
            theta_shadow += 360.0

        return AngleZeroToExclusive360MagnitudeType(
            Angle=theta_shadow,
            Magnitude=numpy.linalg.norm(shadow_prime))

    @property
    def Layover(self):
        # type: () -> AngleZeroToExclusive360MagnitudeType
        """
        AngleMagnitudeType: The layover angle and magnitude.
        """

        layover = self.normal_vector - self.slant_z/(self.slant_z.dot(self.normal_vector))
        theta_layover = numpy.rad2deg(numpy.arctan2(self.row_vector.dot(layover), self.col_vector.dot(layover)))
        if theta_layover < 0:
            theta_layover += 360.0

        return AngleZeroToExclusive360MagnitudeType(
            Angle=theta_layover,
            Magnitude=numpy.linalg.norm(layover))

    @property
    def North(self):
        """Describes the clockwise angle from increasing column direction to north at the center of
        the image.  Angle range: [0, 360)

        Returns
        -------
        float: The north angle.
        """

        lat, lon, _ = ecf_to_geodetic(self.SCP)
        lat_r = numpy.deg2rad(lat)
        lon_r = numpy.deg2rad(lon)
        north = numpy.array(
            [-numpy.sin(lat_r)*numpy.cos(lon_r),
             -numpy.sin(lat_r)*numpy.sin(lon_r),
             numpy.cos(lat_r)])
        north_prime = north - self.slant_z*(north.dot(self.normal_vector)/self.slant_z.dot(self.normal_vector))
        theta_north = numpy.rad2deg(numpy.arctan2(self.row_vector.dot(north_prime), self.col_vector.dot(north_prime)))

        if theta_north < 0:
            theta_north = theta_north + 360.0
        return theta_north

    @property
    def MultiPath(self):
        """
        float: The multipath angle.
        """

        multipath = self.slant_x - self.slant_z*(
            self.slant_x.dot(self.normal_vector)/self.slant_z.dot(self.normal_vector))
        theta_multipath = numpy.rad2deg(numpy.arctan2(self.row_vector.dot(multipath), self.col_vector.dot(multipath)))
        if theta_multipath < 0:
            theta_multipath += 360.0
        return theta_multipath

    @property
    def GroundTrack(self):
        """
        float: The ground track angle.
        """

        track = self.ARPVel - (self.ARPVel.dot(self.normal_vector))*self.normal_vector
        theta_track = numpy.rad2deg(numpy.arctan2(self.row_vector.dot(track), self.col_vector.dot(track)))
        if theta_track < 0:
            theta_track += 360.0
        return theta_track


def _extract_sicd_tx_rcv_pol(str_in):
    """
    Extract the tx and rcv components from the sicd style tx/rcv polarization string.

    Parameters
    ----------
    str_in : str

    Returns
    -------
    str, str
    """

    if str_in is None:
        return 'UNKNOWN', 'UNKNOWN'

    if not isinstance(str_in, str):
        raise TypeError('requires a string type input.')

    if str_in in ['OTHER', 'UNKNOWN']:
        return str_in, str_in

    return str_in.split(':')


class TxRcvPolarizationType(Serializable):
    """
    The transmit/receive polarization information.
    """

    _fields = ('TxPolarization', 'RcvPolarization', 'RcvPolarizationOffset')
    _required = ('TxPolarization', 'RcvPolarization')
    _numeric_format = {'RcvPolarizationOffset': FLOAT_FORMAT}
    # Descriptor
    TxPolarization = StringRegexDescriptor(
        'TxPolarization', TX_POLARIZATION_PATTERN, _required, strict=DEFAULT_STRICT,
        docstring='Transmit polarization type.')  # type: str
    RcvPolarization = StringRegexDescriptor(
        'RcvPolarization', RCV_POLARIZATION_PATTERN, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization type.')  # type: str
    RcvPolarizationOffset = FloatModularDescriptor(
        'RcvPolarizationOffset', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle offset for the receive polarization defined at aperture center.')  # type: float

    def __init__(self, TxPolarization=None, RcvPolarization=None, RcvPolarizationOffset=None, **kwargs):
        """

        Parameters
        ----------
        TxPolarization : str
        RcvPolarization : str
        RcvPolarizationOffset : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPolarization = TxPolarization
        self.RcvPolarization = RcvPolarization
        self.RcvPolarizationOffset = RcvPolarizationOffset
        super(TxRcvPolarizationType, self).__init__(**kwargs)

    @classmethod
    def from_sicd_value(cls, str_in):
        """
        Construct from the sicd style tx/rcv polarization string.

        Parameters
        ----------
        str_in : str

        Returns
        -------
        TxRcvPolarizationType
        """

        tx, rcv = _extract_sicd_tx_rcv_pol(str_in)
        return cls(TxPolarization=tx, RcvPolarization=rcv)


class ProcTxRcvPolarizationType(Serializable):
    """
    The processed transmit/receive polarization.
    """
    _fields = ('TxPolarizationProc', 'RcvPolarizationProc')
    _required = ('TxPolarizationProc', 'RcvPolarizationProc')
    # Descriptor
    TxPolarizationProc = StringRegexDescriptor(
        'TxPolarizationProc', TX_POLARIZATION_PATTERN, _required, strict=DEFAULT_STRICT,
        docstring='Transmit polarization type.')  # type: str
    RcvPolarizationProc = StringRegexDescriptor(
        'RcvPolarizationProc', RCV_POLARIZATION_PATTERN, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization type.')  # type: str

    def __init__(self, TxPolarizationProc=None, RcvPolarizationProc=None, **kwargs):
        """

        Parameters
        ----------
        TxPolarizationProc : str
        RcvPolarizationProc : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPolarizationProc = TxPolarizationProc
        self.RcvPolarizationProc = RcvPolarizationProc
        super(ProcTxRcvPolarizationType, self).__init__(**kwargs)

    @classmethod
    def from_sicd_value(cls, str_in):
        """
        Construct from the sicd style tx/rcv polarization string.

        Parameters
        ----------
        str_in : str

        Returns
        -------
        ProcTxRcvPolarizationType
        """

        tx, rcv = _extract_sicd_tx_rcv_pol(str_in)
        return cls(TxPolarizationProc=tx, RcvPolarizationProc=rcv)


class ExploitationFeaturesCollectionInformationType(Serializable):
    """
    General collection information.
    """

    _fields = (
        'SensorName', 'RadarMode', 'CollectionDateTime', 'LocalDateTime', 'CollectionDuration',
        'Resolution', 'InputROI', 'Polarizations')
    _required = ('SensorName', 'RadarMode', 'CollectionDateTime', 'CollectionDuration')
    _collections_tags = {'Polarizations': {'array': False, 'child_tag': 'Polarization'}}
    _numeric_format = {'CollectionDuration': FLOAT_FORMAT}
    # Descriptor
    SensorName = StringDescriptor(
        'SensorName', _required, strict=DEFAULT_STRICT,
        docstring='The name of the sensor.')  # str
    RadarMode = SerializableDescriptor(
        'RadarMode', RadarModeType, _required, strict=DEFAULT_STRICT,
        docstring='Radar collection mode.')  # type: RadarModeType
    CollectionDateTime = DateTimeDescriptor(
        'CollectionDateTime', _required, strict=DEFAULT_STRICT,
        docstring='Collection date and time defined in Coordinated Universal Time (UTC). The seconds '
                  'should be followed by a Z to indicate UTC.')  # type: numpy.datetime64
    CollectionDuration = FloatDescriptor(
        'CollectionDuration', _required, strict=DEFAULT_STRICT,
        docstring='The duration of the collection (units = seconds).')  # type: float
    Resolution = SerializableDescriptor(
        'Resolution', RangeAzimuthType, _required, strict=DEFAULT_STRICT,
        docstring='Uniformly-weighted resolution (range and azimuth) processed in '
                  'the slant plane.')  # type: Union[None, RangeAzimuthType]
    InputROI = SerializableDescriptor(
        'InputROI', InputROIType, _required, strict=DEFAULT_STRICT,
        docstring='ROI representing portion of input data used to make '
                  'this product.')  # type: Union[None, InputROIType]
    Polarizations = SerializableListDescriptor(
        'Polarizations', TxRcvPolarizationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Transmit and receive polarization(s).')  # type: Union[None, List[TxRcvPolarizationType]]

    def __init__(self, SensorName=None, RadarMode=None, CollectionDateTime=None, LocalDateTime=None,
                 CollectionDuration=None, Resolution=None, Polarizations=None, **kwargs):
        """

        Parameters
        ----------
        SensorName : str
        RadarMode : RadarModeType
        CollectionDateTime : numpy.datetime64|datetime.datetime|datetime.date|str
        LocalDateTime : None|str|datetime.datetime
        CollectionDuration : float
        Resolution : None|RangeAzimuthType|numpy.ndarray|list|tuple
        Polarizations : None|List[TxRcvPolarizationType]
        kwargs
        """

        self._local_date_time = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SensorName = SensorName
        self.RadarMode = RadarMode
        self.CollectionDateTime = CollectionDateTime
        self.CollectionDuration = CollectionDuration
        self.LocalDateTime = LocalDateTime
        self.Resolution = Resolution
        self.Polarizations = Polarizations
        super(ExploitationFeaturesCollectionInformationType, self).__init__(**kwargs)

    @property
    def LocalDateTime(self):
        """None|str:  The local date/time string of the collection. *Optional.*"""
        return self._local_date_time

    @LocalDateTime.setter
    def LocalDateTime(self, value):  # type: (Union[None, str, datetime.datetime]) -> None
        if value is None:
            self._local_date_time = None
            return
        elif isinstance(value, datetime.datetime):
            value = value.isoformat('T')

        if isinstance(value, str):
            self._local_date_time = value
        else:
            logger.error(
                'Attribute LocalDateTime of class ExploitationFeaturesCollectionInformationType\n\t'
                'requires a datetime.datetime or string. Got unsupported type {}.\n\t'
                'Setting value to None.'.format(type(value)))
            self._local_date_time = None

    @classmethod
    def from_sicd(cls, sicd):
        """
        Construct from a sicd element.

        Parameters
        ----------
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesCollectionInformationType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        polarizations = [
            TxRcvPolarizationType.from_sicd_value(entry.TxRcvPolarization)
            for entry in sicd.RadarCollection.RcvChannels]

        return cls(SensorName=sicd.CollectionInfo.CollectorName,
                   RadarMode=RadarModeType(**sicd.CollectionInfo.RadarMode.to_dict()),
                   CollectionDateTime=sicd.Timeline.CollectStart,
                   CollectionDuration=sicd.Timeline.CollectDuration,
                   Resolution=get_collection_resolution(sicd),
                   Polarizations=polarizations)


class ExploitationFeaturesCollectionGeometryType(Serializable):
    """
    Key geometry parameters independent of product processing. All values
    computed at the center time of the full collection.
    """

    _fields = ('Azimuth', 'Slope', 'Squint', 'Graze', 'Tilt', 'DopplerConeAngle', 'Extensions')
    _required = ()
    _collections_tags = {'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {
        'Azimuth': FLOAT_FORMAT, 'Slope': FLOAT_FORMAT, 'Squint': FLOAT_FORMAT, 'Graze': FLOAT_FORMAT,
        'Tilt': FLOAT_FORMAT, 'DopplerConeAngle': FLOAT_FORMAT}
    # Descriptor
    Azimuth = FloatDescriptor(
        'Azimuth', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360.0),
        docstring='Angle clockwise from north indicating the ETP line of sight vector.')  # type: float
    Slope = FloatDescriptor(
        'Slope', _required, strict=DEFAULT_STRICT, bounds=(0.0, 90.0),
        docstring='Angle between the ETP at scene center and the range vector perpendicular to '
                  'the direction of motion.')  # type: float
    Squint = FloatModularDescriptor(
        'Squint', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle from the ground track to platform velocity vector at nadir. '
                  'Left-look is positive, right-look is negative.')  # type: float
    Graze = FloatDescriptor(
        'Graze', _required, strict=DEFAULT_STRICT, bounds=(0.0, 90.0),
        docstring='Angle between the ETP and the line of sight vector.')  # type: float
    Tilt = FloatModularDescriptor(
        'Tilt', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle between the ETP and the cross range vector. '
                  'Also known as the twist angle.')  # type: float
    DopplerConeAngle = FloatDescriptor(
        'DopplerConeAngle', _required, strict=DEFAULT_STRICT, bounds=(0.0, 180.0),
        docstring='The angle between the velocity vector and the radar line-of-sight vector. '
                  'Also known as the slant plane squint angle.')  # type: float
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Azimuth=None, Slope=None, Squint=None, Graze=None, Tilt=None,
                 DopplerConeAngle=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Azimuth : None|float
        Slope : None|float
        Squint : None|float
        Graze : None|float
        Tilt : None|float
        DopplerConeAngle : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Azimuth = Azimuth
        self.Slope = Slope
        self.Squint = Squint
        self.Graze = Graze
        self.Tilt = Tilt
        self.DopplerConeAngle = DopplerConeAngle
        self.Extensions = Extensions
        super(ExploitationFeaturesCollectionGeometryType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator

        Returns
        -------
        ExploitationFeaturesCollectionGeometryType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(_exp_calc_text.format(type(calculator)))

        return cls(Azimuth=calculator.AzimuthAngle,
                   Slope=calculator.SlopeAngle,
                   Graze=calculator.GrazeAngle,
                   Tilt=calculator.TiltAngle,
                   DopplerConeAngle=calculator.DopplerConeAngle,
                   Squint=calculator.SquintAngle)


class ExploitationFeaturesCollectionPhenomenologyType(Serializable):
    """
    Phenomenology related to both the geometry and the final product processing.
    All values computed at the center time of the full collection.
    """

    _fields = ('Shadow', 'Layover', 'MultiPath', 'GroundTrack', 'Extensions')
    _required = ()
    _collections_tags = {'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {'MultiPath': FLOAT_FORMAT, 'GroundTrack': FLOAT_FORMAT}
    # Descriptor
    Shadow = SerializableDescriptor(
        'Shadow', AngleZeroToExclusive360MagnitudeType, _required, strict=DEFAULT_STRICT,
        docstring='The phenomenon where vertical objects occlude radar '
                  'energy.')  # type: Union[None, AngleZeroToExclusive360MagnitudeType]
    Layover = SerializableDescriptor(
        'Layover', AngleZeroToExclusive360MagnitudeType, _required, strict=DEFAULT_STRICT,
        docstring='The phenomenon where vertical objects appear as ground objects with '
                  'the same range/range rate.')  # type: Union[None, AngleZeroToExclusive360MagnitudeType]
    MultiPath = FloatDescriptor(
        'MultiPath', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360.0),
        docstring='This is a range dependent phenomenon which describes the energy from a '
                  'single scatter returned to the radar via more than one path and results '
                  'in a nominally constant direction in the ETP.')  # type: Union[None, float]
    GroundTrack = FloatDescriptor(
        'GroundTrack', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360.0),
        docstring='Counter-clockwise angle from increasing column direction to ground track '
                  'at the center of the image.')  # type: Union[None, float]
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Shadow=None, Layover=None, MultiPath=None, GroundTrack=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Shadow : None|AngleZeroToExclusive360MagnitudeType|numpy.ndarray|list|tuple
        Layover : None|AngleZeroToExclusive360MagnitudeType|numpy.ndarray|list|tuple
        MultiPath : None|float
        GroundTrack : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Shadow = Shadow
        self.Layover = Layover
        self.MultiPath = MultiPath
        self.GroundTrack = GroundTrack
        self.Extensions = Extensions
        super(ExploitationFeaturesCollectionPhenomenologyType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator

        Returns
        -------
        ExploitationFeaturesCollectionPhenomenologyType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(_exp_calc_text.format(type(calculator)))
        return cls(Shadow=calculator.Shadow, Layover=calculator.Layover,
                   MultiPath=calculator.MultiPath, GroundTrack=calculator.GroundTrack)


class CollectionType(Serializable):
    """
    Metadata regarding one of the input collections.
    """
    _fields = ('Information', 'Geometry', 'Phenomenology', 'identifier')
    _required = ('Information', 'identifier')
    _set_as_attribute = ('identifier', )
    # Descriptor
    Information = SerializableDescriptor(
        'Information', ExploitationFeaturesCollectionInformationType, _required, strict=DEFAULT_STRICT,
        docstring='General collection information.')  # type: ExploitationFeaturesCollectionInformationType
    Geometry = SerializableDescriptor(
        'Geometry', ExploitationFeaturesCollectionGeometryType, _required, strict=DEFAULT_STRICT,
        docstring='Key geometry parameters independent of product '
                  'processing.')  # type: Union[None, ExploitationFeaturesCollectionGeometryType]
    Phenomenology = SerializableDescriptor(
        'Phenomenology', ExploitationFeaturesCollectionPhenomenologyType, _required, strict=DEFAULT_STRICT,
        docstring='Phenomenology related to both the geometry and the final '
                  'product processing.')  # type: Union[None, ExploitationFeaturesCollectionPhenomenologyType]
    identifier = StringDescriptor(
        'identifier', _required, strict=DEFAULT_STRICT,
        docstring='The exploitation feature identifier.')  # type: str

    def __init__(self, Information=None, Geometry=None, Phenomenology=None, identifier=None, **kwargs):
        """

        Parameters
        ----------
        Information : ExploitationFeaturesCollectionInformationType
        Geometry : None|ExploitationFeaturesCollectionGeometryType
        Phenomenology : None|ExploitationFeaturesCollectionPhenomenologyType
        identifier : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Information = Information
        self.Geometry = Geometry
        self.Phenomenology = Phenomenology
        self.identifier = identifier
        super(CollectionType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator, sicd):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator
        sicd : SICDType

        Returns
        -------
        CollectionType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(_exp_calc_text.format(type(calculator)))

        return cls(identifier=sicd.CollectionInfo.CoreName,
                   Information=ExploitationFeaturesCollectionInformationType.from_sicd(sicd),
                   Geometry=ExploitationFeaturesCollectionGeometryType.from_calculator(calculator),
                   Phenomenology=ExploitationFeaturesCollectionPhenomenologyType.from_calculator(calculator))


class ExploitationFeaturesProductType(Serializable):
    """
    Metadata regarding the product.
    """

    _fields = ('Resolution', 'Ellipticity', 'Polarizations', 'North', 'Extensions')
    _required = ('Resolution', 'Ellipticity', 'Polarizations')
    _collections_tags = {
        'Polarizations': {'array': False, 'child_tag': 'Polarization'},
        'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {'Ellipticity': FLOAT_FORMAT, 'North': FLOAT_FORMAT}
    # Descriptor
    Resolution = SerializableDescriptor(
        'Resolution', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Uniformly-weighted resolution projected into the Earth Tangent '
                  'Plane (ETP).')  # type: RowColDoubleType
    Ellipticity = FloatDescriptor(
        'Ellipticity', _required, strict=DEFAULT_STRICT,
        docstring="Ellipticity of the 2D-IPR at the ORP, measured in the *Earth Geodetic "
                  "Tangent Plane (EGTP)*. Ellipticity is the ratio of the IPR ellipse's "
                  "major axis to minor axis.")  # type: float
    Polarizations = SerializableListDescriptor(
        'Polarizations', ProcTxRcvPolarizationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Describes the processed transmit and receive polarizations for the '
                  'product.')  # type: List[ProcTxRcvPolarizationType]
    North = FloatDescriptor(
        'North', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360.0),
        docstring='Clockwise angle from increasing column direction to north at the center'
                  'of the image.')  # type: float
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Resolution=None, Ellipticity=None, Polarizations=None,
                 North=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Resolution : RowColDoubleType|numpy.ndarray|list|tuple
        Ellipticity : float
        Polarizations : List[ProcTxRcvPolarizationType]
        North : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Resolution = Resolution
        self.Ellipticity = Ellipticity
        self.Polarizations = Polarizations
        self.North = North
        self.Extensions = Extensions
        super(ExploitationFeaturesProductType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator, sicd):
        """
        Construct from a sicd element.

        Parameters
        ----------
        calculator : ExploitationCalculator
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesProductType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        rho_rg_slant, rho_az_slant = get_collection_resolution(sicd)
        rho_row_etp, rho_col_etp = calculator.get_ground_plane_resolution(rho_rg_slant, rho_az_slant)
        ellipticity = rho_row_etp/rho_col_etp if rho_row_etp >= rho_col_etp else rho_col_etp/rho_row_etp

        return cls(Resolution=(rho_row_etp, rho_col_etp),
                   Ellipticity=ellipticity,
                   Polarizations=[
                       ProcTxRcvPolarizationType.from_sicd_value(sicd.ImageFormation.TxRcvPolarizationProc), ],
                   North=calculator.North)


class ExploitationFeaturesType(Serializable):
    """
    Computed metadata regarding the collect.
    """
    _fields = ('Collections', 'Products')
    _required = ('Collections', 'Products')
    _collections_tags = {
        'Collections': {'array': False, 'child_tag': 'Collection'},
        'Products': {'array': False, 'child_tag': 'Product'}}
    # Descriptor
    Collections = SerializableListDescriptor(
        'Collections', CollectionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[CollectionType]
    Products = SerializableListDescriptor(
        'Products', ExploitationFeaturesProductType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[ExploitationFeaturesProductType]

    def __init__(self, Collections=None, Products=None, **kwargs):
        """

        Parameters
        ----------
        Collections : List[CollectionType]
        Products : List[ExploitationFeaturesProductType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Collections = Collections
        self.Products = Products
        super(ExploitationFeaturesType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd, row_vector, col_vector):
        """
        Construct from a sicd element.

        Parameters
        ----------
        sicd : SICDType|List[SICDType]
        row_vector : numpy.ndarray
        col_vector : numpy.ndarray

        Returns
        -------
        ExploitationFeaturesType
        """
        if isinstance(sicd, (list, tuple)):
            collections = []
            feats = []
            for i, entry in sicd:
                calculator = ExploitationCalculator.from_sicd(entry, row_vector, col_vector)
                collections.append(CollectionType.from_calculator(calculator, entry))
                feats.append(ExploitationFeaturesProductType.from_calculator(calculator, entry))
            return cls(Collections=collections, Products=feats)

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))
        calculator = ExploitationCalculator.from_sicd(sicd, row_vector, col_vector)
        return cls(
            Collections=[CollectionType.from_calculator(calculator, sicd), ],
            Products=[ExploitationFeaturesProductType.from_calculator(calculator, sicd)])



"""
Establish base expected functionality for digital elevation model handling.
"""

import numpy
from typing import List

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class DEMInterpolator(object):
    """
    Abstract DEM class presenting base required functionality.
    """

    def get_elevation_hae(self, lat, lon, block_size=50000):
        """
        Get the elevation value relative to the WGS-84 ellipsoid.

        Parameters
        ----------
        lat : numpy.ndarray|list|tuple|int|float
        lon : numpy.ndarray|list|tuple|int|float
        block_size : int|None
            If `None`, then the entire calculation will proceed as a single block.
            Otherwise, block processing using blocks of the given size will be used.
            The minimum value used for this is 50,000, and any smaller value will be
            replaced with 50,000. Default is 50,000.

        Returns
        -------
        numpy.ndarray
            the elevation relative to the WGS-84 ellipsoid.
        """

        raise NotImplementedError

    def get_elevation_geoid(self, lat, lon, block_size=50000):
        """
        Get the elevation value relative to the geoid.

        Parameters
        ----------
        lat : numpy.ndarray|list|tuple|int|float
        lon : numpy.ndarray|list|tuple|int|float
        block_size : int|None
            If `None`, then the entire calculation will proceed as a single block.
            Otherwise, block processing using blocks of the given size will be used.
            The minimum value used for this is 50,000, and any smaller value will be
            replaced with 50,000. Default is 50,000.

        Returns
        -------
        numpy.ndarray
            the elevation relative to the geoid
        """

        raise NotImplementedError

    def get_max_hae(self, lat_lon_box=None):
        """
        Get the maximum dem value with respect to HAE, which should be assumed
        **approximately** correct. This may possibly be with respect to some
        Area of Interest.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            None or any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """

        raise NotImplementedError

    def get_min_hae(self, lat_lon_box=None):
        """
        Get the minimum dem value with respect to HAE, which should be assumed
        **approximately** correct. This may possibly be with respect to some
        Area of Interest.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            None or any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """

        raise NotImplementedError

    def get_max_geoid(self, lat_lon_box=None):
        """
        Get the maximum dem value with respect to the geoid, which should be assumed
        **approximately** correct. This may possibly be with respect to some
        Area of Interest.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            None or any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """

        raise NotImplementedError

    def get_min_geoid(self, lat_lon_box=None):
        """
        Get the minimum dem value with respect to geoid, whihc should be assumed
        **approximately** correct. This may possibly be with respect to some
        Area of Interest.

        Parameters
        ----------
        lat_lon_box : None|numpy.ndarray
            None or any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """

        raise NotImplementedError


class DEMList(object):
    """
    Abstract class for creating a searchable list of applicable DEM files of a
    given type.
    """

    def get_file_list(self, lat_lon_box):
        """
        This will return the list of files associated with covering the
        `lat_lon_box` using a DEM. Extraneous files (i.e. with region not overlapping
        the provided box) should NOT be returned, and files should be returned in
        order of preference.

        .. Note: It should be considered the user's responsibility to ensure
            that necessary DEM files will be found by this methodology, and
            regions lacking DEM file(s) should be assumed to have elevation
            at WGS-84 mean sea level.

        Parameters
        ----------
        lat_lon_box : numpy.ndarray|list|tuple
            The bounding box of the form `[lat min, lat max, lon min, lon max]`.

        Returns
        -------
        List[str]
        """

        raise NotImplementedError



"""
Helper classes and methods for Fourier processing schemes.
"""

__classification__ = "UNCLASSIFIED"
__author__ = 'Thomas McCullough'

import logging
from typing import Union

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.processing.ortho_rectify import FullResolutionFetcher

# NB: the below are intended as common imports from other locations
#   leave them here, even if unused
import numpy
import scipy
if scipy.__version__ < '1.4':
    # noinspection PyUnresolvedReferences
    from scipy.fftpack import fft, ifft, fftshift, ifftshift
else:
    # noinspection PyUnresolvedReferences
    from scipy.fft import fft, ifft, fftshift, ifftshift

logger = logging.getLogger(__name__)


class FFTCalculator(FullResolutionFetcher):
    """
    Base Fourier processing calculator class.

    This is intended for processing schemes where full resolution is required
    along the processing dimension, so sub-sampling along the processing
    dimension does not decrease the amount of data which must be fetched.
    """

    __slots__ = (
        '_platform_direction', '_fill')

    def __init__(
            self,
            reader: Union[str, SICDTypeReader],
            dimension: int = 0,
            index: int = 0,
            block_size: Union[None, int, float] = 50):
        """

        Parameters
        ----------
        reader : str|SICDTypeReader
            Input file path or reader object, which must be of sicd type.
        dimension : int
            The dimension over which to split the sub-aperture.
        index : int
            The sicd index to use.
        block_size : int
            The approximate processing block size to fetch, given in MB. The
            minimum value for use here will be 1.
        """

        self._platform_direction = None  # set with the index setter
        self._fill = None  # set implicitly with _set_fill()
        super(FFTCalculator, self).__init__(reader, dimension=dimension, index=index, block_size=block_size)

    @property
    def dimension(self) -> int:
        """
        int: The dimension along which to perform the color subaperture split.
        """

        return self._dimension

    @dimension.setter
    def dimension(self, value):
        value = int(value)
        if value not in [0, 1]:
            raise ValueError('dimension must be 0 or 1, got {}'.format(value))
        self._dimension = value
        self._set_fill()

    @property
    def index(self) -> int:
        """
        int: The index of the reader.
        """

        return self._index

    @index.setter
    def index(self, value):
        super(FFTCalculator, self)._set_index(value)

        if self._sicd.SCPCOA is None or self._sicd.SCPCOA.SideOfTrack is None:
            logger.warning(
                'The sicd object at index {} has unpopulated SCPCOA.SideOfTrack.\n\t'
                'Defaulting to "R", which may be incorrect.')
            self._platform_direction = 'R'
        else:
            self._platform_direction = self._sicd.SCPCOA.SideOfTrack
        self._set_fill()

    @property
    def fill(self) -> float:
        """
        float: The fill factor for the fourier processing.
        """

        return self._fill

    def _set_fill(self):
        self._fill = None
        if self._dimension is None:
            return
        if self._index is None:
            return

        if self.dimension == 0:
            try:
                fill = 1.0/(self.sicd.Grid.Row.SS*self.sicd.Grid.Row.ImpRespBW)
            except (ValueError, AttributeError, TypeError):
                fill = 1.0
        else:
            try:
                fill = 1.0/(self.sicd.Grid.Col.SS*self.sicd.Grid.Col.ImpRespBW)
            except (ValueError, AttributeError, TypeError):
                fill = 1.0
        self._fill = max(1.0, float(fill))

    def __getitem__(self, item) -> numpy.ndarray:
        """
        Fetches the processed data based on the input slice.

        Parameters
        ----------
        item

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError


def _validate_fft_input(array: numpy.ndarray) -> None:
    """
    Validate the fft input.

    Parameters
    ----------
    array : numpy.ndarray

    Returns
    -------
    None
    """

    if not isinstance(array, numpy.ndarray):
        raise TypeError('array must be a numpy array')
    if not numpy.iscomplexobj(array):
        raise ValueError('array must have a complex data type')
    if array.ndim != 2:
        raise ValueError('array must be a two-dimensional array. Got shape {}'.format(array.shape))


def _determine_direction(
        sicd: SICDType, dimension: int) -> int:
    """
    Determine the default sign for the fft.

    Parameters
    ----------
    sicd : SICDType
    dimension : int

    Returns
    -------
    int
    """

    sgn = None
    if dimension == 0:
        try:
            sgn = sicd.Grid.Row.Sgn
        except AttributeError:
            pass
    elif dimension == 1:
        try:
            sgn = sicd.Grid.Col.Sgn
        except AttributeError:
            pass
    else:
        raise ValueError('dimension must be one of 0 or 1.')
    return -1 if sgn is None else sgn


def fft_sicd(array: numpy.ndarray, dimension: int, sicd: SICDType) -> numpy.ndarray:
    """
    Apply the forward one-dimensional forward fft to data associated with the
    given sicd along the given dimension/axis, in accordance with the sign
    populated in the SICD structure (default is -1).

    Parameters
    ----------
    array : numpy.ndarray
        The data array, which must be two-dimensional and complex.
    dimension : int
        Must be one of 0, 1.
    sicd : SICDType
        The associated SICD structure.

    Returns
    -------
    numpy.ndarray
    """

    sgn = _determine_direction(sicd, dimension)
    return fft(array, axis=dimension) if sgn < 0 else ifft(array, axis=dimension)


def ifft_sicd(array: numpy.ndarray, dimension: int, sicd: SICDType) -> numpy.ndarray:
    """
    Apply the inverse one-dimensional fft to data associated with the given sicd
    along the given dimension/axis.

    Parameters
    ----------
    array : numpy.ndarray
        The data array, which must be two-dimensional and complex.
    dimension : int
        Must be one of 0, 1.
    sicd : SICDType
        The associated SICD structure.

    Returns
    -------
    numpy.ndarray
    """

    sgn = _determine_direction(sicd, dimension)
    return ifft(array, axis=dimension) if sgn < 0 else fft(array, axis=dimension)


def fft2_sicd(array: numpy.ndarray, sicd: SICDType) -> numpy.ndarray:
    """
    Apply the forward two-dimensional fft (i.e. both axes) to data associated with
    the given sicd.

    Parameters
    ----------
    array : numpy.ndarray
        The data array, which must be two-dimensional and complex.
    sicd : SICDType
        The associated SICD structure.

    Returns
    -------
    numpy.ndarray
    """

    return fft_sicd(fft_sicd(array, 0, sicd), 1, sicd)


def ifft2_sicd(array: numpy.ndarray, sicd: SICDType) -> numpy.ndarray:
    """
    Apply the inverse two-dimensional fft (i.e. both axes) to data associated with
    the given sicd.

    Parameters
    ----------
    array : numpy.ndarray
        The data array, which must be two-dimensional and complex.
    sicd : SICDType
        The associated SICD structure.

    Returns
    -------
    numpy.ndarray
    """

    return ifft_sicd(ifft_sicd(array, 0, sicd), 1, sicd)



"""
This module provide utilities for reading essentially Compensated Phase History Data.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import os
from typing import BinaryIO, Callable, Union

from sarpy.io.general.base import SarpyIOError, check_for_openers
from sarpy.io.general.utils import is_file_like
from sarpy.io.phase_history.base import CPHDTypeReader


###########
# Module variables
_openers = []
_parsed_openers = False


def register_opener(open_func: Callable) -> None:
    """
    Provide a new opener.

    Parameters
    ----------
    open_func : Callable
        This is required to be a function which takes a single argument (file name).
        This function should return a sarpy.io.phase_history.base.CPHDTypeReader instance
        if the referenced file is viable for the underlying type, and None otherwise.

    Returns
    -------
    None
    """

    if not callable(open_func):
        raise TypeError('open_func must be a callable')
    if open_func not in _openers:
        _openers.append(open_func)


def parse_openers() -> None:
    """
    Automatically find the viable openers (i.e. :func:`is_a`) in the various modules.
    """

    global _parsed_openers
    if _parsed_openers:
        return
    _parsed_openers = True

    check_for_openers('sarpy.io.phase_history', register_opener)


def open_phase_history(file_name: Union[str, BinaryIO]) -> CPHDTypeReader:
    """
    Given a file, try to find and return the appropriate reader object.

    Parameters
    ----------
    file_name : str|BinaryIO

    Returns
    -------
    CPHDTypeReader

    Raises
    ------
    SarpyIOError
    """

    if (not is_file_like(file_name)) and (not os.path.exists(file_name)):
        raise SarpyIOError('File {} does not exist.'.format(file_name))
    # parse openers, if not already done
    parse_openers()
    # see if we can find a reader though trial and error
    for opener in _openers:
        reader = opener(file_name)
        if reader is not None:
            return reader

    # If for loop completes, no matching file format was found.
    raise SarpyIOError('Unable to determine phase history image format.')



"""
The CompressionType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union

import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    FloatArrayDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT


class J2KSubtype(Serializable):
    """
    The Jpeg 2000 subtype.
    """
    _fields = ('NumWaveletLevels', 'NumBands', 'LayerInfo')
    _required = ('NumWaveletLevels', 'NumBands')
    _collections_tags = {'LayerInfo': {'array': True, 'child_tag': 'Bitrate', 'size_attribute': 'numLayers'}}
    _numeric_format = {'LayerInfo': FLOAT_FORMAT}
    # Descriptor
    NumWaveletLevels = IntegerDescriptor(
        'NumWaveletLevels', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: int
    NumBands = IntegerDescriptor(
        'NumBands', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: int
    LayerInfo = FloatArrayDescriptor(
        'LayerInfo', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Original Layer Information. This is an array of bit rate target associated with each '
                  'layer. It may happen that the bit rate was not achieved due to data characteristics. '
                  '**Note -** for JPEG 2000 numerically loss-less quality, the bit rate for the final layer is '
                  'an expected value, based on performance.')  # type: Union[None, numpy.ndarray]

    def __init__(self, NumWaveletLevels=None, NumBands=None, LayerInfo=None, **kwargs):
        """

        Parameters
        ----------
        NumWaveletLevels : int
        NumBands : int
        LayerInfo : None|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumWaveletLevels = NumWaveletLevels
        self.NumBands = NumBands
        self.LayerInfo = LayerInfo
        super(J2KSubtype, self).__init__(**kwargs)


class J2KType(Serializable):
    """
    Jpeg 2000 parameters.
    """

    _fields = ('Original', 'Parsed')
    _required = ('Original', )
    # Descriptor
    Original = SerializableDescriptor(
        'Original', J2KSubtype, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: J2KSubtype
    Parsed = SerializableDescriptor(
        'Parsed', J2KSubtype, _required, strict=DEFAULT_STRICT,
        docstring='Conditional fields that exist only for parsed images.')  # type: Union[None, J2KSubtype]

    def __init__(self, Original=None, Parsed=None, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Original = Original
        self.Parsed = Parsed
        super(J2KType, self).__init__(**kwargs)


class CompressionType(Serializable):
    """
    Contains information regarding any compression that has occurred to the image data.
    """

    _fields = ('J2K', )
    _required = ('J2K', )
    # Descriptor
    J2K = SerializableDescriptor(
        'J2K', J2KType, _required, strict=DEFAULT_STRICT,
        docstring='Block describing details of JPEG 2000 compression.')  # type: J2KType

    def __init__(self, J2K=None, **kwargs):
        """

        Parameters
        ----------
        J2K : J2KType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.J2K = J2K
        super(CompressionType, self).__init__(**kwargs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PIATGBType(TREElement):
    def __init__(self, value):
        super(PIATGBType, self).__init__()
        self.add_field('TGTUTM', 's', 15, value)
        self.add_field('PIATGAID', 's', 15, value)
        self.add_field('PIACTRY', 's', 2, value)
        self.add_field('PIACAT', 's', 5, value)
        self.add_field('TGTGEO', 's', 15, value)
        self.add_field('DATUM', 's', 3, value)
        self.add_field('TGTNAME', 's', 38, value)
        self.add_field('PERCOVER', 's', 3, value)
        self.add_field('TGTLAT', 's', 10, value)
        self.add_field('TGTLON', 's', 11, value)


class PIATGB(TREExtension):
    _tag_value = 'PIATGB'
    _data_type = PIATGBType
